//@ts-check
"use strict";

/**
@typedef Dialog
@type {{
  id: string,
  speaker: string,
  options?: string[],
}}
 */

/**
@typedef DialogState
@type {{
  pixiRoot: PIXI.Container;
  background: PIXI.NineSlicePlane;
  speakerText: PIXI.Text;
  contentText: PIXI.Text;
  currentDialog: Dialog
}}
 */

/** @returns DialogState */
function createDialog() {
  const pixiRoot = new PIXI.Container();
  pixiRoot.scale.set(inverseTileSize, inverseTileSize);

  const background = new PIXI.NineSlicePlane(tex("inGame/hud/dialog"), 6, 6, 6, 14);
  background.width = 192;
  pixiRoot.addChild(background);

  const speakerText = new PIXI.Text("SPEAKER:", new PIXI.TextStyle({
    fontSize: 28, fontFamily: "Arvo", fontWeight: "bold",
    fill: "#fff", strokeThickness: 4 * app.renderer.height / gameHeight, stroke: "#513a43", lineJoin: "bevel"
  }));
  speakerText.position.set(6, -8);
  speakerText.scale.set(1 / 3);
  pixiRoot.addChild(speakerText);

  const contentText = new PIXI.Text("Henlo", new PIXI.TextStyle({
    fontSize: 28, fontFamily: "Arvo",
    fill: "#fff", strokeThickness: 4 * app.renderer.height / gameHeight, stroke: "#513a43", lineJoin: "bevel",
    wordWrap: true, wordWrapWidth: (background.width - 12) * 3
  }));
  contentText.position.set(6, 6);
  contentText.scale.set(1 / 3);
  pixiRoot.addChild(contentText);

  /** @type DialogState */
  let res = {
    pixiRoot,
    background,
    speakerText,
    contentText,
    currentDialog: null,
  };

  return res;
}

/**
 * @param {NPC} npc
 * @returns {Dialog}
 */
function initialDialog(npc) {
  ////////////////////////////////////
  // King Gudul & his dog
  ////////////////////////////////////
  if (npc.desc.id === "gudul") {
    if (justCompletedQuest("gudulDogRescue")) {
      return {
        id: "gudulFoundMyDog",
        speaker: npc.id,
      }
    } else if (canSuggestQuest("gudulDogRescue")) {
      return {
        id: "gudulLostDog",
        speaker: npc.id,
        options: ["lookForDog", "bye"],
      };
    } else if (isQuestInProgress("gudulDogRescue")) {
      return {
        id: "gudulFoundDogYet",
        speaker: npc.id,
      };
    }
  }
  if (npc.desc.id === "pup") {
    if (npc.following) {
      return {
        id: "woofQuestion",
        speaker: npc.id,
      };
    } else {
      return {
        id: "woof",
        speaker: npc.id,
      };
    }
  }
  ////////////////////////////////////
  // Captain
  ////////////////////////////////////
  if (npc.desc.id === "captain") {
    if (canSuggestQuest("shipRepair")) {
      return {
        id: "captainBrokenShip",
        options: ["willLook", "bye"],
        speaker: npc.id,
      };
    } else if (justCompletedQuest("shipRepair")) {
      return {
        id: "captainThanksPlanks",
        speaker: npc.id,
      }
    } else if (isQuestInProgress("shipRepair")) {
      let options = [];
      if (countFollowers() > 0) {
        options.push("embark");
      }
      options.push("bye");
      return {
        id: "captainWantsPlanks",
        speaker: npc.id,
        options,
      }
    } else {
      let options = [];
      options.push("sail");
      if (countFollowers() > 0) {
        options.push("embark");
      }
      options.push("bye");
      return {
        id: "captainHello",
        speaker: npc.id,
        options,
      }
    }
  }

  ////////////////////////////////////
  // Mom & bro
  ////////////////////////////////////
  if (npc.desc.id === "bro") {
    if (canSuggestQuest("saveMom")) {
      return {
        id: "broLetsGo",
        options: ["go", "bye"],
        speaker: npc.id,
      }
    } else if (isQuestInProgress("saveMom")) {
      return {
        id: "broSavingMom",
        speaker: npc.id,
      }
    } else if (isQuestDone("saveMom")) {
      addFollower(findNPC("bro"));
      return {
        id: "broMomAlreadySaved",
        speaker: npc.id,
      }
    } else {
      return null;
    }
  }
  if (npc.desc.id === "mom") {
    if (!wasQuestAlreadyDone("saveMom")) {
      applyQuestOutcome("saveMom");
      return {
        id: "momHi",
        options: ["yes"],
        speaker: npc.id,
      };
    }
  }

  ////////////////////////////////////
  // Grandma
  ////////////////////////////////////
  if (npc.desc.id === "grandma") {
    if (canSuggestQuest("grandmaMinion")) {
      return {
        id: "grandmaWants1",
        speaker: npc.id,
      }
    } else if (justCompletedQuest("grandmaMinion")) {
      return {
        id: "grandmaHappy",
        speaker: npc.id,
      }
    } else if (isQuestInProgress("grandmaMinion")) {
      return {
        id: "grandmaStillWants",
        speaker: npc.id,
      }
    }
  }

  ////////////////////////////////////
  // Greedy bro
  ////////////////////////////////////
  if (npc.desc.id === "broody") {
    if (canSuggestQuest("greedyBro")) {
      return {
        id: "greedyBroWants1",
        options: ["yes", "no"],
        speaker: npc.id,
      };
    } else if (justCompletedQuest("greedyBro")) {
      return {
        id: "greedyBroThanksForBucks",
        speaker: npc.id,
      };
    } else if (isQuestInProgress("greedyBro")) {
      return {
        id: "greedyBroWheresMyMoney",
        speaker: npc.id,
      };
    }
  }

  return null;
}

/**
 * @param {Dialog} prev - previous dialog
 * @param {string} option - picked option (null if just 'continue')
 * @returns {Dialog}
 */
function nextDialog(prev, option) {
  let next = nextDialogReal(prev, option);
  console.table({ prev, option, next });
  return next;
}

/**
 * @param {Dialog} prev - previous dialog
 * @param {string} option - picked option (null if just 'continue')
 * @returns {Dialog}
 */
function nextDialogReal(prev, option) {
  ////////////////////////////////////
  // King Gudul & his dog
  ////////////////////////////////////
  switch (prev.id) {
    case "woof": {
      return {
        id: "foundPup1",
        speaker: null,
      };
    }
    case "woofQuestion": {
      return null;
    }
    case "foundPup1": {
      let pup = findNPC("pup");
      if (pup && addFollower(pup)) {
        return {
          id: "foundPup2",
          speaker: null,
        };
      } else {
        return {
          id: "tooManyFriends",
          speaker: null,
        };
      }
    }
    case "foundPup2": {
      return null;
    }
    case "gudulLostDog": {
      let gudul = findNPC("gudul");
      console.log({ option, gudul });
      if (option === "lookForDog" && gudul) {
        return {
          id: "gudulThanksForLookingForMyDog",
          speaker: gudul.id,
        }
      }
      return null;
    }
    case "gudulThanksForLookingForMyDog":
      startQuest("gudulDogRescue");
      return null;
  }

  ////////////////////////////////////
  // Captain
  ////////////////////////////////////
  switch (prev.id) {
    case "captainBrokenShip": {
      switch (option) {
        case "willLook":
          startQuest("shipRepair");
          return null;
        case "bye":
          return null;
      }
      break;
    }
    case "captainWantsPlanks": {
      switch (option) {
        case "embark": {
          doEmbark();
          return {
            id: "captainAllAboard",
            speaker: prev.speaker,
          };
        }
        case "bye": {
          return null;
        }
      }
    }
    case "captainThanksPlanks": {
      let options = [];
      options.push("sail");
      if (countFollowers() > 0) {
        options.push("embark");
      }
      options.push("bye");
      return {
        id: "captainHello",
        speaker: prev.speaker,
        options,
      }
    }
    case "captainHello": {
      switch (option) {
        case "sail": {
          doEmbark();
          doWin();
          return null;
        }
        case "bye": { return null; }
        case "embark": {
          doEmbark();
          return {
            id: "captainAllAboard",
            speaker: prev.speaker,
          };
        }
      }
    }
  }

  ////////////////////////////////////
  // Mom & bro
  ////////////////////////////////////
  switch (prev.id) {
    case "broLetsGo": {
      switch (option) {
        case "go": {
          addFollower(findNPC("bro"));
          startQuest("saveMom");
          return null;
        }
        case "bye": {
          return null;
        }
      }
      break;
    }
    case "momHi": {
      return null;
    }
  }

  ////////////////////////////////////
  // Grandma
  ////////////////////////////////////
  if (prev.id === "grandmaWants1") {
    return {
      id: "grandmaWants2",
      speaker: prev.speaker,
      options: ["yes"],
    }
  }
  if (prev.id === "grandmaWants2") {
    if (option === "yes") {
      startQuest("grandmaMinion");
    }
    return null;
  }

  ////////////////////////////////////
  // Greedy bro
  ////////////////////////////////////
  if (prev.id === "greedyBroWants1") {
    if (option === "yes") {
      startQuest("greedyBro");
      return {
        id: "greedyBroThanksForLooking",
        speaker: prev.speaker,
      }
    }
    return null;
  } else if (prev.id === "greedyBroThanksForBucks") {
    return null;
  }

  return null;
}

/** @param {Dialog} dialog */
function setCurrentDialog(dialog) {
  if (!inGame) {
    return;
  }
  inGame.hud.isCrafting = false;
  inGame.dialog.currentDialog = dialog;
}

/** @param {number} dt */
function updateDialog(dt) {
  const { dialog } = inGame;
  const { pixiRoot, background, speakerText, contentText, currentDialog } = dialog;

  if (currentDialog) {
    pixiRoot.visible = true;

    let speaker = inGame.activeWorld.npcs.find((npc) => npc.id === currentDialog.speaker);
    let targetX;
    let targetY;
    if (speaker) {
      targetX = speaker.pos.x;
      targetY = speaker.pos.y;
    } else {
      targetX = inGame.baxton.pos.x;
      targetY = inGame.baxton.pos.y;
    }

    let speakerName = speaker ? getText(`npcs.${speaker.desc.id}.name`) : "Baxton";
    speakerText.text = speakerName.toUpperCase() + ":";

    /** @type string[] */
    let lines = [];
    lines.push(getText(`dialogs.${currentDialog.id}.line`));
    lines.push("");
    if (isEmpty(currentDialog.options)) {
      lines.push(getText(`dialogMeta.continue`));
    } else {
      for (let i = 0; i < currentDialog.options.length; i++) {
        let option = currentDialog.options[i];
        lines.push(`${i + 1}. ` + getText(`dialogs.${currentDialog.id}.options.${option}`));
      }
    }
    contentText.text = lines.join("\n");

    background.height = contentText.height + 6 + 14;
    targetX -= 3;
    targetY -= background.height / tileSize + 1;
    pixiRoot.position.set(targetX, targetY);

  } else {
    pixiRoot.visible = false;
    contentText.text = "";
  }

}

/**
 * @param {number | null} optionNumber
 */
function selectDialogOption(optionNumber) {
  const { currentDialog } = inGame.dialog;
  if (!currentDialog) {
    return;
  }

  if (optionNumber === null) {
    if (!isEmpty(currentDialog.options)) {
      if (currentDialog.options.length == 1) {
        setCurrentDialog(nextDialog(currentDialog, currentDialog.options[0]));
      } else {
        // can't just "continue", gotta pick an option
        return;
      }
    } else {
      setCurrentDialog(nextDialog(currentDialog, null));
    }
  } else {
    if (isEmpty(currentDialog.options)) {
      // have to press space to continue
      return;
    }

    if (!currentDialog.options[optionNumber]) {
      // option doesn't exist
      return;
    }

    setCurrentDialog(nextDialog(currentDialog, currentDialog.options[optionNumber]));
  }
}

function doEmbark() {
  const { stats } = inGame;
  let oldCount = countSavedNPCs();

  for (const k of Object.keys(stats.followers)) {
    stats.savedNPCs[k] = true;
    delete stats.followers[k];
    // @ts-ignore
    let npc = findNPC(k);
    destroyNpc(npc);
  }

  let newCount = countSavedNPCs();
  if (newCount > oldCount) {
    notify("huge", getText(`questMeta.totalFriendsSaved`) + newCount);
  }
}

/** @return {number} */
function countSavedNPCs() {
  const { stats } = inGame;
  return Object.keys(stats.savedNPCs).length;
}

function doWin() {
  /** @type {EndState} */
  let endState = {
    savedFolks: countSavedNPCs(),
  };
  leaveInGame(() => {
    enterWin(endState);
  })
}